// react:
import { 
// hooks:
useRef, useEffect, } from 'react';
// cssfn:
import { 
// writes css in javascript:
rule, states, style, vars, cssVars, } from '@cssfn/core'; // writes css in javascript
// reusable-ui utilities:
import { 
// hooks:
useTriggerRender, useEvent, } from '@reusable-ui/hooks'; // react helper hooks
// reusable-ui features:
import { 
// hooks:
usesAnimation, } from '@reusable-ui/animation'; // animation stuff of UI
const [excitableVars] = cssVars();
{
    const { animationRegistry: { registerFilter, registerAnim } } = usesAnimation();
    registerFilter(excitableVars.filter);
    registerAnim(excitableVars.anim);
}
// parent is     `.excited` -or-  current is     `.excited`:
export const ifExcited = (styles) => rule(':is(.excited&, &.excited)', styles); // specificityWeight = 1 + (parent's specificityWeight)
// parent is not `.excited` -and- current is not `.excited`:
export const ifNotExcited = (styles) => rule(':where(&):not(:is(.excited&, &.excited))', styles); // specificityWeight = 1 + (parent's specificityWeight)
/**
 * Adds a capability of UI to highlight itself to attract user's attention.
 * @param config  A configuration of `excitableRule`.
 * @returns An `ExcitableStuff` represents an excitable state.
 */
export const usesExcitable = (config) => {
    return {
        excitableRule: () => style({
            // animation states:
            ...states([
                ifExcited({
                    ...vars({
                        [excitableVars.filter]: config?.filterExcite,
                        [excitableVars.anim]: config?.animExcite,
                    }),
                }),
            ]),
        }),
        excitableVars,
    };
};
export const useToggleExcitable = (props) => {
    // props:
    const { 
    // states:
    excited, onExcitedChange, } = props;
    /*
     * the state is excited/normal based on [controllable excited]
     */
    const excitedFn = (excited /*controllable*/ ?? false);
    // states:
    // local storages without causing to (re)render, we need to manual control the (re)render event:
    /**
     * `true`  => was excited
     * `false` => was normal
     */
    const wasExcited = useRef(excitedFn);
    // manually controls the (re)render event:
    const [triggerRender] = useTriggerRender();
    const asyncTriggerRender = useRef(undefined);
    useEffect(() => {
        // cleanups:
        return () => {
            // cancel out previously asyncTriggerRender (if any):
            if (asyncTriggerRender.current)
                cancelIdleCallback(asyncTriggerRender.current);
        };
    }, []); // runs once on startup
    if (wasExcited.current !== excitedFn) { // change detected => apply the change & start animating
        const continueToRun = excitedFn && (wasExcited.current === null);
        if (continueToRun) {
            // cancel out previously asyncTriggerRender (if any):
            if (asyncTriggerRender.current)
                cancelIdleCallback(asyncTriggerRender.current);
            // wait until the non-excited `<Component>` has been applied by browser ui, then re-render the excited `<Component>`
            asyncTriggerRender.current = requestIdleCallback(() => {
                wasExcited.current = excitedFn; // remember the last change
                triggerRender(); // re-render the excited `<Component>`
            });
        }
        else {
            wasExcited.current = excitedFn; // remember the last change
        } // if
    } // if
    // handlers:
    const handleAnimationEnd = useEvent((event) => {
        // conditions:
        if (event.target !== event.currentTarget)
            return; // ignores bubbling
        if (!/((?<![a-z])(excite)|(?<=[a-z])(Excite))(?![a-z])/.test(event.animationName))
            return; // ignores animation other than excite[Foo] or booExcite[Foo]
        // clean up finished animation
        const continueToRun = wasExcited.current;
        wasExcited.current = null; // mark the animation need to restart
        Promise.resolve().then(() => {
            onExcitedChange?.({ excited: false }); // request to stop
        });
        if (continueToRun) {
            triggerRender(); // need to restart the animation
        } // if
    });
    return {
        excited: wasExcited.current,
        class: wasExcited.current ? 'excited' : null,
        handleAnimationEnd,
    };
};
//#endregion excitable
